#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2024 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from Bio.Seq import Seq
from pathlib import Path
import numpy as np

from pimento.bin.pimento_utils import (
    get_read_count,
    compute_windowed_base_conservation,
    build_list_of_base_counts,
    fetch_read_substrings,
)
from pimento.bin.thresholds import CONSENSUS_BASE_THRESHOLD, MAX_READ_COUNT


def choose_cutoff_for_single_strand(
    input_fastq: Path, cutoff_list: list[int], rev: bool = False
) -> None:
    """
    Assess inflection point list, selecting one for automatic primer trimming.

    Takes as input a fastq file and a list of inflection points generated by "find_cutoffs.py".
    Computes the average conservation of bases before inflection point and after.
    Gets the difference in avg. conservation between the pre- and post- points.
    Selects the inflection point with the maximum difference as the cutoff.
    If an inf point has a similar difference and is earlier than the max, we make a 'conservative' choice and
    replace it with the earlier cutoff.

    Returns the cutoff point and the consensus sequence 'forming' the automatically predicted primer
    """

    start_confs = []  # pre-inf point conservations
    end_confs = []  # post-inf point conservations
    start_cons_lens = []  # list for storing lengths of pre-inflection point sequences
    cons_seq_list = []  # list for storing consensus sequences pre-inflection points

    do_not_include_list = [
        i + 5 for i in cutoff_list
    ]  # ignore conservation of inflection point in calculation

    read_count = get_read_count(input_fastq)  # get readcount from fastq

    max_line_count = 0
    if read_count > MAX_READ_COUNT:
        max_line_count = MAX_READ_COUNT

    for start in cutoff_list:  # Looping through the pre-inflection point substrings
        substring_len = (
            start + 4
        )  # length of pre-inf substrings is inflection point + 4

        read_substring_count_dict = fetch_read_substrings(
            input_fastq, substring_len, rev=rev, max_line_count=max_line_count
        )  # get base count dict
        base_counts = build_list_of_base_counts(
            read_substring_count_dict, substring_len
        )  # list of base conservation dicts for substrings
        base_conservation, cons_seq = compute_windowed_base_conservation(
            base_counts,
            read_count,
            CONSENSUS_BASE_THRESHOLD,
            do_not_include_list,
            max_line_count=max_line_count,
        )  # get list of max base conservations for each index
        # also get consensus sequence
        cons_seq_list.append(cons_seq)
        start_confs.append(np.mean(base_conservation))
        start_cons_lens.append(len(cons_seq))

    for i, end in enumerate(
        cutoff_list
    ):  # Looping through the post-inflection point substrings
        substring_len = end + 5  # length of pre-inf substrings is inflection point + 5
        subs_len = start_cons_lens[i]  # length of respective pre-inf point sequence

        read_substring_count_dict = fetch_read_substrings(
            input_fastq, subs_len, rev, substring_len, max_line_count=max_line_count
        )
        base_counts = build_list_of_base_counts(read_substring_count_dict, subs_len)
        base_conservation, cons_seq = compute_windowed_base_conservation(
            base_counts,
            read_count,
            CONSENSUS_BASE_THRESHOLD,
            do_not_include_list,
            subs_len,
            max_line_count=max_line_count,
        )

        end_confs.append(np.mean(base_conservation))

    diff_res = [
        start_confs[i] - end_confs[i] for i in range(len(start_confs))
    ]  # get differences between pre- and -post avg conservation values
    diff_res_sorted = sorted(
        diff_res, reverse=True
    )  # sort differences from highest to lowest

    ini_max_res = diff_res_sorted[0]  # maximum differences
    curr_max_index = diff_res.index(ini_max_res)  # index of maximum differences

    for res in diff_res_sorted[1:]:  # Loop through the rest of the differences
        curr_res_index = np.where(diff_res == res)[0][0]

        index_diff = cutoff_list[curr_max_index] - cutoff_list[curr_res_index]

        # if difference between the max and the current is negligible and the index of the current is earlier then..
        if ini_max_res - res < 0.05 and (index_diff <= 3 and index_diff > 0):
            curr_max_index = (
                curr_res_index  # replace the selected index with the current one
            )

    cutoff = cutoff_list[curr_max_index] + 5  # cutoff is the inflection point index + 5
    primer = cons_seq_list[
        curr_max_index
    ]  # grab the correct consensus sequence as primer

    # if the requested strand is reverse..
    if rev:
        primer = str(
            Seq(primer).complement()
        )  # ..get the complement of consensus sequence

    return cutoff, primer



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2025 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict

import numpy as np

from pimento.bin.thresholds import MAX_CUTOFF_WINDOW_START, MIN_CUTOFF_WINDOW_START


def find_bcv_inflection_points(bcv_df):
    """
    Find inflection points from an bcv_df file output by "generate_bcv.py"

    Takes the list of average base conservations and gets the derivative of the curve
    Keep any points of the derivative where value is above the 80th percentile

    Outputs a dictionary with key-val pairs where vals are lists:
        'strand' -> strand list
        'inf_point' -> inf_point list

    """

    inf_point_dict = defaultdict(list)
    start_indices = [int(i) for i in bcv_df.columns.tolist()]

    for i in range(len(bcv_df)):  # Loop through both possible strands of the bcv_df
        strand = bcv_df.index[i]
        props = bcv_df.iloc[i].tolist()
        props = [-val for val in props]

        prop_diff = np.diff(props) / np.diff(start_indices)  # Get the derivative
        infl_points = np.where(prop_diff > np.percentile(prop_diff, 80))[
            0
        ]  # Grab points above 80th percentile

        for ind in infl_points:
            inf_point = start_indices[ind]

            if (
                inf_point < MIN_CUTOFF_WINDOW_START
                or inf_point > MAX_CUTOFF_WINDOW_START
            ):  # Rule to facilitate results - won't accept
                continue  # points below index 10 or above index 20
                # 10 means a cutoff of 15 and 20 a cutoff of 25
                # literature points to no primers existing that are
                # shorter or bigger  than these lengths

            inf_point_dict["strand"].append(strand)
            inf_point_dict["inf_point"].append(inf_point)

    return inf_point_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2025 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from pathlib import Path

import pandas as pd
import numpy as np

from pimento.bin.pimento_utils import (
    get_read_count,
    compute_windowed_base_conservation,
    build_list_of_base_counts,
    fetch_read_substrings,
)
from pimento.bin.thresholds import MAX_READ_COUNT, BCV_WINDOW_SIZE


def generate_bcv_for_single_strand(path: Path, rev: bool = False) -> defaultdict[float]:
    """
    Generate base-conservation vectors in a stepwise and windowed manner for a fastq file.

    For a continuous range of starting indices (2 to 25), generate substrings of window size of 5 bases.
    Calculate the average conservation of the most common base at each index of a window.
    The resulting list of base-conservations can be considered a conservation vector and used to
    identify inflection points where the conservation suddenly changes.

    Output a dictionary where:
        key -> an index starting point e.g. base 10
        val -> the avg conservation of the most common base for the substring window from base 10 to 15 (inclusive)
    """

    res_dict = defaultdict(float)
    start_range = range(2, 25, 1)  # Range of starting indices

    for start in start_range:

        read_count = get_read_count(
            path, file_type="fastq"
        )  # get read count for fastq file

        max_line_count = 0
        if read_count > MAX_READ_COUNT:
            max_line_count = MAX_READ_COUNT

        read_substring_count_dict = fetch_read_substrings(
            path, BCV_WINDOW_SIZE, rev, start, max_line_count
        )  # get substring count dict
        base_counts = build_list_of_base_counts(
            read_substring_count_dict, BCV_WINDOW_SIZE
        )  # list of base conservation dicts for substring
        base_conservation, cons_seq = compute_windowed_base_conservation(
            base_counts, read_count, max_line_count=max_line_count
        )  # get list of max base conservations for each index

        res_dict[start] = np.mean(base_conservation)  # compute the mean

    return res_dict


def write_bcv_output(
    fwd_out: defaultdict[float] = {}, rev_out: defaultdict[float] = {}
):
    """
    Generate Pandas dataframe out of bcv dictionary.

    Output looks like this (when both F and R are requested):
        2	3	4
    F	0.7814975041597337	0.8736772046589019	0.9434276206322796
    R	0.9010981697171381	0.9082861896838601	0.90369384359401

    Columns are the starting indices. Row labels are the strand.
    """

    final_res_dict = defaultdict(list)
    df_ind = []

    # Check if fwd strand was requested
    if fwd_out:
        [final_res_dict[key].append(fwd_out[key]) for key in fwd_out.keys()]
        df_ind.append("F")

    # Check if rev strand was requested
    if rev_out:
        [final_res_dict[key].append(rev_out[key]) for key in rev_out.keys()]
        df_ind.append("R")

    res_df = pd.DataFrame.from_dict(final_res_dict)
    res_df.index = df_ind

    return res_df



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2024 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
import numpy as np

from pimento.bin.pimento_utils import (
    get_read_count,
    compute_windowed_base_conservation,
    build_list_of_base_counts,
    fetch_read_substrings,
)

from pimento.bin.thresholds import ATP_WINDOW_SIZE, ATP_PREFIX_LENGTH


def atp_in_this_sample(input_fastq: Path, rev: bool = False) -> bool:
    """
    Predict the presence of primers based on windows of base conservation.

    Takes a fastq file as input. Extracts proportion of most common base for the first 100 bases.
    Computes the a threshold (Q3 - 0.15) based on this proportion and counts the number of bases below
    it in windows of 10 bases.
    If at least one of the first two windows contains at most one such a base, then the presence
    of a primer is flagged as true. A primer is also flagged as true if the combined count
    of bases below Q3 is at most 4.

    The output of this function is a boolean flag:
        True if a primer was identified
        False if a primer was not identified
    """

    read_count = get_read_count(
        input_fastq, file_type="fastq"
    )  # Get read count for fastq file

    read_substring_count_dict = fetch_read_substrings(
        input_fastq, ATP_PREFIX_LENGTH, rev=rev
    )  # substring dict where key is the substring and value is the count
    base_counts = build_list_of_base_counts(
        read_substring_count_dict, ATP_PREFIX_LENGTH
    )  # list of base conservation dicts for substrings
    base_conservation, cons_seq = compute_windowed_base_conservation(
        base_counts, read_count
    )  # get list of max base conservations for each index

    # Counter that will reset to 0 every 10 bases
    window_count = 0
    # Will append the window count to this list every 10 bases
    window_count_list = []
    # Compute Q3-based threshold
    max_cons = np.quantile(base_conservation, 0.75)
    threshold = max_cons - 0.15

    if max_cons < 0.75:
        threshold = 0.75
    # Immediately return false (no primer) if the max conservation is less than 0.6
    if max_cons < 0.6:
        return False

    # Loop through every base
    for counter, val in enumerate(base_conservation):
        if (
            counter % ATP_WINDOW_SIZE == 0 and counter != 0
        ):  # After looping through a window..
            window_count_list.append(window_count)  # ..append window count
            window_count = 0  # ..reset window count

        if (
            val < threshold
        ):  # If the conservation at i is less than threshold, increment count for the window
            window_count += 1

    primer_flag = False  # Initialise primer flag as false

    if (
        1 in window_count_list[:2] or 0 in window_count_list[:2]
    ):  # If window count is at most 1 of first two windows...
        primer_flag = True  # ..primer flag is true
    elif (
        sum(window_count_list[:2]) <= 4
    ):  # If sum of window counts of the first two windows is at most 4..
        primer_flag = True  # ..primer flag is true

    return primer_flag


def write_atp_output(results: tuple[bool, bool], output_prefix: str) -> None:
    """
    Save primer presence flags into output .txt file.

    1: primer exists
    0: primer doesn't exist

    First line will be the forward strand
    Second line will be the reverse strand
    """

    with open(f"{output_prefix}_general_primer_out.txt", "w") as fw:
        fw.write(f"{results[0]}\n")
        fw.write(f"{results[1]}\n")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2025 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# used by bin.amplicon_utils.fetch_read_substrings
MAX_READ_COUNT = 300_000
# used by bin.amplicon_utils.build_cons_seq
CONSENSUS_BASE_THRESHOLD = 0.80

# used by pimento.standard_primer_strategy
MIN_STD_PRIMER_THRESHOLD = 0.60

# used by bin.standard_primer_matching.run_primer_matching_once
STD_PRIMER_READ_PREFIX_LENGTH = 50

# used by bin.are_there_primers.atp_in_this_sample
ATP_WINDOW_SIZE = 10
ATP_PREFIX_LENGTH = 100

# used by bin.generate_bcv.generate_bcv_for_single_strand
BCV_WINDOW_SIZE = 5

# used by bin.find_cutoffs.find_bcv_inflection_points
MIN_CUTOFF_WINDOW_START = 10
MAX_CUTOFF_WINDOW_START = 20



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2025 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

AMBIGUOUS_BASES_DICT = {
    "R": "[AG]",
    "Y": "[CT]",
    "S": "[GC]",
    "W": "[AT]",
    "K": "[GT]",
    "M": "[AC]",
    "B": "[CGT]",
    "D": "[AGT]",
    "H": "[ACT]",
    "V": "[ACG]",
    "N": "[ACTG]",
}

AMBIGUOUS_BASES_DICT_REV = {
    "A,G": "R",
    "C,T": "Y",
    "C,G": "S",
    "A,T": "W",
    "G,T": "K",
    "A,C": "M",
    "C,G,T": "B",
    "A,G,T": "D",
    "A,C,T": "H",
    "A,C,G": "V",
    "A,C,G,T": "N",
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2025 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict, Counter
from pathlib import Path
import logging
import pyfastx

from pimento.bin.regex_ambiguous_bases import (
    AMBIGUOUS_BASES_DICT,
    AMBIGUOUS_BASES_DICT_REV,
)
from pimento.bin.thresholds import CONSENSUS_BASE_THRESHOLD

logging.basicConfig(level=logging.DEBUG)


def get_read_count(read_path: Path, file_type: str = "fastq") -> int:
    """
    Get the read count of a FASTQ or FASTA file.

    :param read_path: The path to the FASTQ or FASTA file.
    :type read_path: Path
    :param fasta_type: The type of the file, either "fastq" or "fasta". Defaults to "fastq".
    :type fasta_type: str
    :return: The number of reads in the file.
    :rtype: int
    :raises ValueError: If the file type is not supported or the read count is not a positive integer.
    """
    read_count = 0

    if file_type == "fasta":
        fasta = pyfastx.Fasta(str(read_path), build_index=False)
        read_count = sum(1 for _ in fasta)
    elif file_type == "fastq":
        fastq = pyfastx.Fastq(str(read_path), build_index=False)
        read_count = sum(1 for _ in fastq)
    else:
        raise ValueError(
            f"Invalid file_type {file_type}, it needs to be either 'fasta' or 'fastq'"
        )

    if read_count <= 0:
        raise ValueError(f"Read count is not a positive integer: {read_count}")

    return read_count


def compute_windowed_base_conservation(
    cons_list: list,
    read_count: int,
    cons_threshold: float = CONSENSUS_BASE_THRESHOLD,
    do_not_include: list = [],
    counter_start: int = 1,
    max_line_count: int = 0,
) -> tuple[str, list[float]]:
    """
    Generate consensus sequence using a list of base conservation dictionaries most likely
    generated by the `build_list_of_base_counts()` function.
    Also returns a list containing the conservation value of the most conserved base at every
    position in the list of base conservation dictionaries.
    """

    cons_seq = ""
    base_conservation = []

    for counter, count_dict in enumerate(cons_list, counter_start):
        max_count = 0
        cons_dict = defaultdict(float)

        if len(count_dict) == 1 and "N" in count_dict:
            max_prop = 0.0
            cons_seq = "N"
            continue

        if counter in do_not_include:
            continue

        for base, count in count_dict.items():
            if base not in ("A", "T", "C", "G"):
                continue

            if max_line_count == 0:
                cons_dict[base] = count / read_count
            else:
                cons_dict[base] = count / max_line_count

            if count > max_count:
                max_count = count

        try:
            if max_line_count == 0:
                max_prop = max_count / read_count
            else:
                max_prop = max_count / max_line_count

            cons_bases = []
            curr_prop = 0.0
            sorted_cons_dict = dict(
                sorted(cons_dict.items(), key=lambda x: x[1], reverse=True)
            )

            for base, prop in sorted_cons_dict.items():
                cons_bases.append(base)
                curr_prop += prop
                if curr_prop >= cons_threshold:
                    break

            cons_bases = sorted(cons_bases)

            if len(cons_bases) == 1:
                cons_seq += cons_bases[0]
            else:
                amb_string = ",".join(cons_bases)
                amb_base = AMBIGUOUS_BASES_DICT_REV[amb_string]
                cons_seq += amb_base

        except ZeroDivisionError:
            max_prop = 0.0

        base_conservation.append(max_prop)

    return base_conservation, cons_seq


def primer_regex_query_builder(primer):
    """
    Takes an input nucleotide sequence that can contain IUPAC ambiguous codes
    Returns a string formatted as a regex query that considers the different
    potential bases valid at a position with am abiguity code.
    """

    query = ""

    for char in primer:
        if char in ("A", "C", "T", "G"):
            query += char
        else:
            query += str(AMBIGUOUS_BASES_DICT[char])

    query = f"(.*{query}){{e<=2}}"

    return query


def build_list_of_base_counts(
    read_substring_count_dict: dict, prefix_len: int
) -> list[defaultdict[int]]:
    """
    Generate list of dictionaries of base conservation counts for substring output (fetch_read_substrings)
    e.g. [{'A':220, 'C':14}, {'T':234}, ....] for every base position
    """

    substring_cons_list = []

    for i in range(prefix_len):
        index_base_dict = defaultdict(int)
        for substring in read_substring_count_dict.keys():
            if len(substring) < prefix_len:
                continue
            base = substring[i]
            index_base_dict[base] += read_substring_count_dict[substring]
        substring_cons_list.append(index_base_dict)

    return substring_cons_list


def fetch_read_substrings(
    input_fastq: Path,
    prefix_len: int,
    rev: bool = False,
    start: int = 1,
    max_line_count: int = 0,
) -> dict:
    """
    Generates the most common prefix sequences along with their counts in a fastq file.
    Outputs dictionary containing counts for each generated substring in the fastq.
    """

    selected_lines = []

    fastq = pyfastx.Fastq(str(input_fastq), build_index=False)

    for read in fastq:
        sequence = read[1]  # the read sequence is the second element
        if not rev:
            selected_lines.append(sequence[start - 1 : start + prefix_len - 1])
        else:
            rev_sequence = sequence[::-1]
            selected_lines.append(rev_sequence[start - 1 : start + prefix_len - 1])
        if max_line_count != 0:
            if len(selected_lines) > max_line_count:
                break

    sequence_counts = Counter(selected_lines)
    substring_count_dict = dict(
        sorted(sequence_counts.items(), key=lambda x: x[1], reverse=True)
    )

    return substring_count_dict



from pathlib import Path
import importlib.metadata
from importlib.resources import files

import click
import pandas as pd
import pyfastx
from rich.console import Console
from rich import print

from pimento.bin.standard_primer_matching import (
    get_primer_props,
    parse_std_primers,
    write_std_output,
)
from pimento.bin.are_there_primers import atp_in_this_sample, write_atp_output
from pimento.bin.generate_bcv import generate_bcv_for_single_strand, write_bcv_output
from pimento.bin.find_cutoffs import find_bcv_inflection_points
from pimento.bin.choose_primer_cutoff import choose_cutoff_for_single_strand

from pimento.bin.thresholds import MIN_STD_PRIMER_THRESHOLD

console = Console()
__version__ = importlib.metadata.version("mi-pimento")
DEFAULT_STD_PRIMERS_PATH = files("pimento.standard_primers").joinpath("")


@click.group()
@click.version_option(__version__)
def cli():
    console.log(
        "[bold grey74]Starting new [bold green]PI[/bold green][bold red]MENTO[/bold red] run![bold grey74]\n"
    )
    pass


@cli.command(
    "std",
    options_metavar="-i <fastq/fastq.gz> -p <primers_dir> -o <output_prefix>",
    short_help="Perform the standard primer strategy for primer inference",
)
@click.option(
    "-i",
    "--input_fastq",
    required=True,
    help="Input fastq file to perform primer inference on.",
    type=click.Path(exists=True, path_type=Path, dir_okay=False),
)
@click.option(
    "-p",
    "--primers_dir",
    required=True,
    help="Input directory containing the standard primer library. Default uses the PIMENTO standard primer library.",
    type=click.Path(exists=True, path_type=Path, file_okay=False),
    default=Path(DEFAULT_STD_PRIMERS_PATH),
)
@click.option(
    "-m",
    "--minimum_primer_threshold",
    help="The minimum proportion of reads a standard primer has to be present\
in to be considered in inference. Default value of 0.60.",
    type=float,
    default=MIN_STD_PRIMER_THRESHOLD,
)
@click.option(
    "-o", "--output_prefix", required=True, help="Prefix to output file.", type=str
)
@click.option(
    "--merged",
    default=False,
    is_flag=True,
    help="Flag for running the standard primer strategy in `merged` mode, \
which is necessary if the input sequence file is made up of merged paired-end \
or single-end reads. Specifically, it will reverse the substrings that are searched\
for reverse primers, and use the complement of said reverse primers.",
)
def standard_primer_strategy(
    input_fastq: Path,
    primers_dir: Path,
    minimum_primer_threshold: float,
    output_prefix: str,
    merged: bool,
) -> None:
    """Runs the standard primer matching strategy for primer inference.
    A library of standard primers will be searched in input reads using fuzzy regex to identify
    primers with high degrees of proportions. The best single, or pair, of primers found will be
    outputted into a FASTA file (provided any were indeed found).
    PIMENTO comes with its own standard primer library which will be used as a default. Users can
    however give their own custom input library, or extend the default library to suit their needs.

    :param input_fastq: The input FASTQ file.
    :type input_fastq: Path
    :param primers_dir: Directory containing FASTA files of primer sequences for the library.
    The only format requirement is that forward strand primer names in the FASTA headers
    finish with the character `F`, and vice versa `R` for reverse strand primers. See PIMENTO's
    standard library for examples.
    :type primers_dir: Path
    :param minimum_primer_threshold: The minimum matching proportion threshold for a primer to be considered.
    The current default value for this threshold is a proportion of 0.60 of reads. Users can customise this value.
    :type minimum_primer_threshold: float
    :param output_prefix: The prefix to be used on output files.
    :type output_prefix: str
    """
    print(
        "[bold grey74]Running [bold green]standard primer strategy[/bold green].[/bold grey74]"
    )
    print(
        f"[bold grey74]Input FASTQ file: [bold green]{input_fastq}[/bold green][/bold grey74]"
    )
    print(
        f"[bold grey74]Standard primer library: [bold green]{primers_dir}[/bold green][/bold grey74]"
    )
    print(
        f"[bold grey74]Output prefix: [bold green]{output_prefix}[/bold green][/bold grey74]"
    )
    print("")

    with console.status("[bold yellow]Loading standard primer library..."):
        std_primer_dict_regex, std_primer_dict, primer_count = parse_std_primers(
            primers_dir, merged
        )  # Parse std primer library into dictionaries
        console.log(
            "[bold green]Loading standard primer library :white_check_mark:[/bold green]\n"
        )

    print(
        f"[bold grey74][bold green]{primer_count} primers[/bold green] loaded[/bold grey74]"
    )

    with console.status("[bold yellow]Searching for standard primers..."):
        results = get_primer_props(
            std_primer_dict_regex, input_fastq, minimum_primer_threshold, merged
        )  # Find all the std primers in the input and select most common
        console.log(
            "[bold green]Searching for standard primers :white_check_mark:[/bold green]\n"
        )

    std_primers_fasta, std_primers_info = write_std_output(
        results, output_prefix, std_primer_dict, merged
    )

    if results:
        primer_fasta = pyfastx.Fasta(str(std_primers_fasta), build_index=False)
        print("[bold grey74]Found standard library primers:[/bold grey74]")
        for primer in primer_fasta:
            print(f"[bold green]>{primer[0]}[/bold green]")
            print(f"[bold red]{primer[1]}[/bold red]")
        print("")
        print(
            f"[bold grey74]Final primers found saved to [bold green]{std_primers_fasta}[/bold green][/bold grey74]"
        )
        print(
            f"[bold grey74]Proportion information saved to [bold green]{std_primers_info}[/bold green][/bold grey74]"
        )
        print(
            "[bold grey74]All proportions saved to [bold green]all_standard_primer_proportions.txt[/bold green]"
            "[/bold grey74]"
        )
    else:
        print("No standard library primers found in reads!")
        print(
            "[bold grey74]Log of proportions for all primers found in "
            "[bold green]all_standard_primer_proportions.txt[/bold green][/bold grey74]"
        )


@cli.command(
    "are_there_primers",
    options_metavar="-i <fastq/fastq.gz> -o <output_prefix>",
    short_help="Predict whether primers are present in the input reads",
)
@click.option(
    "-i",
    "--input_fastq",
    required=True,
    help="Input fastq file to predict the presence of primers for.",
    type=click.Path(exists=True, path_type=Path, dir_okay=False),
)
@click.option(
    "-o", "--output_prefix", required=True, help="Prefix to output file.", type=str
)
def are_there_primers(input_fastq: Path, output_prefix: str) -> None:
    """Checks for the presence of primers in input reads.
    Using patterns of base-conservation values for input reads, this function makes a prediction
    for whether there are primers at either ends of reads. The flag for whether primers are detected
    (1 for yes, 0 for no) for both strands is saved to an output file.

    :param input_fastq: The input FASTQ file.
    :type input_fastq: Path
    :param output_prefix: The prefix to be used on output files.
    :type output_prefix: str
    """
    fwd_primer_flag = atp_in_this_sample(
        input_fastq
    )  # Check for general primers in fwd
    rev_primer_flag = atp_in_this_sample(
        input_fastq, rev=True
    )  # Check for general primers in rev

    fwd_status = "0"
    rev_status = "0"
    # Flag for primer presence: 1 for yes 0 for no
    if fwd_primer_flag:
        print("Forward primer detected!")
        fwd_status = 1
    else:
        print("No forward primer detected")
    if rev_primer_flag:
        print("Reverse primer detected!")
        rev_status = 1
    else:
        print("No reverse primer detected")

    write_atp_output(
        (fwd_status, rev_status), output_prefix
    )  # Save primer flags to .txt file


@cli.command(
    "gen_bcv",
    options_metavar="-i <fastq/fastq.gz> -st [FR/F/R] -o <output_prefix>",
    short_help="Generate the base-conservation vector(s) (BCV)",
)
@click.option(
    "-i",
    "--input_fastq",
    required=True,
    help="Input fastq file to generate the BCV for.",
    type=click.Path(exists=True, path_type=Path, dir_okay=False),
)
@click.option(
    "-st",
    "--strand",
    help="The strand(s) to generate a BCV for. Values can be either F, R, or FR for both.",
    type=click.Choice(["FR", "F", "R"]),
    required=True,
)
@click.option(
    "-o", "--output_prefix", required=True, help="Prefix to output file.", type=str
)
def generate_base_conservation_vector(
    input_fastq: Path, strand: str, output_prefix: str
) -> Path:
    """Generates base-conservation vector(s) for input fastq files.
    To be used by the primer cutoff inference strategy, this function computes the base-conservation
    vector for input reads. Users can choose to perform this computation for either forward or reverse
    strands of the reads, or both.

    :param input_fastq: The input FASTQ file.
    :type input_fastq: Path
    :param strand: The strand(s) to perform primer inference for. Values can be either F, R, or FR for both.
    :type strand: str
    :param output_prefix: The prefix to be used on output files.
    :type output_prefix: str
    :return: TSV file containing the base-conservation vector.
    :rtype: Path
    """

    with console.status("[bold yellow]Generating base-conservation vector..."):

        res_df = ""

        # TODO: match-case statement is python 3.10>. We are currently locking the version
        # at version 3.9. The day we bump the version we should replace these if statements
        # with a match-case block.

        if strand == "FR":
            fwd_bcv = generate_bcv_for_single_strand(input_fastq)
            rev_bcv = generate_bcv_for_single_strand(input_fastq, rev=True)
            res_df = write_bcv_output(fwd_bcv, rev_bcv)
        elif strand == "F":
            fwd_bcv = generate_bcv_for_single_strand(input_fastq)
            res_df = write_bcv_output(fwd_bcv)
        elif strand == "R":
            rev_bcv = generate_bcv_for_single_strand(input_fastq, rev=True)
            res_df = write_bcv_output(rev_out=rev_bcv)

        # Save resulting dataframe to a tsv file
        res_df.to_csv(f"{output_prefix}_bcv.tsv", sep="\t")

    console.log(
        "[bold green]Generating base-conservation vector :white_check_mark:[/bold green]"
    )

    return f"{output_prefix}_bcv.tsv"


@cli.command(
    "find_cutoffs",
    options_metavar="-i <bcv.tsv> -o <output_prefix>",
    short_help="Find potential cutoffs using a BCV output.",
)
@click.option(
    "-i",
    "--input_bcv",
    required=True,
    help="Input BCV file to identify potential cutoffs for.",
    type=click.Path(exists=True, path_type=Path, dir_okay=False),
)
@click.option(
    "-o", "--output_prefix", required=True, help="Prefix to output file.", type=str
)
def find_potential_cutoffs(input_bcv: Path, output_prefix: str) -> Path:
    """Find the potential cutoffs in a base-conservation vector output file.
    By computing negative inflection points in the base-conservation vectors,
    this function identifies potential cutoff points to be considered as primer cutoff
    points in the next and final step of the primer cutoff inference strategy. These potential
    cutoffs are saved to a TSV file.

    :param input_bcv: TSV file containing the base-conservation vector.
    :type input_bcv: Path
    :param output_prefix: The prefix to be used on output files.
    :type output_prefix: str
    :return: TSV file containing the potential cutoff points to consider.
    :rtype: Path
    """
    with console.status("[bold yellow]Finding potential cutoffs..."):

        bcv_df = pd.read_csv(input_bcv, sep="\t", index_col=0)  # Read mcp_df
        inf_point_dict = find_bcv_inflection_points(
            bcv_df
        )  # Generate inflection points dict

        if len(inf_point_dict) > 0:  # If the inf_point_dict isn't empty..
            inf_point_df = pd.DataFrame.from_dict(
                inf_point_dict
            )  # .. turn it into a dataframe
            inf_point_df.to_csv(
                f"{output_prefix}_cutoffs.tsv", sep="\t", index=False
            )  # ..save it to a .tsv file

        else:  # If it is empty..
            fw = open(f"{output_prefix}_cutoffs.tsv", "w")  # ..make an empty file
            fw.close()

    console.log("[bold green]Finding potential cutoffs :white_check_mark:[/bold green]")

    return Path(f"{output_prefix}_cutoffs.tsv")


@cli.command(
    "choose_primer_cutoff",
    options_metavar="-i <fastq/fastq.gz> -p <cutoffs.tsv> -o <output_prefix>",
    short_help="Choose the optimal primer cutoff point.",
)
@click.option(
    "-i",
    "--input_fastq",
    required=True,
    help="Input fastq file to choose the optimal primer cutoff point for.",
    type=click.Path(exists=True, path_type=Path, dir_okay=False),
)
@click.option(
    "-p",
    "--primer_cutoffs",
    required=True,
    help="File containing the potential cutoff points to choose from.",
    type=click.Path(exists=True, path_type=Path, dir_okay=False),
)
@click.option(
    "-o", "--output_prefix", required=True, help="Prefix to output file.", type=str
)
def choose_primer_cutoff(
    input_fastq: Path, primer_cutoffs: Path, output_prefix: str
) -> Path:
    """Choose the optimal primer cutoff point from an input set.
    Using patterns of base-conservation in reads and a set of potential cutoff points
    to choose from, this function makes a prediction for the "optimal" cutoff point
    to be used for primer inference. This function then generates a consensus sequence
    using the reads and the chosen cutoff point to be used as the final primer sequence.
    Performed for both strands, the final primer sequences are output into a FASTA file.

    :param input_fastq: The input FASTQ file.
    :type input_fastq: Path
    :param primer_cutoffs: TSV file containing the potential cutoff points to consider.
    :type primer_cutoffs: Path
    :param output_prefix: The prefix to be used on output files.
    :type output_prefix: str
    :return: The output FASTA file containing the inferred primer sequences.
    :rtype: Path
    """
    with console.status("[bold yellow]Choosing optimal primer cutoff point..."):

        cutoffs_df = pd.read_csv(primer_cutoffs, sep="\t")

        f_slice = cutoffs_df[cutoffs_df.strand == "F"]  # get forward inflection points
        r_slice = cutoffs_df[cutoffs_df.strand == "R"]  # get reverse inflection points
        r_slice = r_slice.reset_index(drop=True)

        f_cutoff = ""
        r_cutoff = ""
        f_primer = ""
        r_primer = ""

        if not f_slice.empty:  # if there is a forward inflection point..
            cutoff_list = f_slice.inf_point.tolist()
            f_cutoff, f_primer = choose_cutoff_for_single_strand(
                input_fastq, cutoff_list
            )  # .. assess and select

        if not r_slice.empty:  # if there is a reverse inflection point..
            cutoff_list = r_slice.inf_point.tolist()
            r_cutoff, r_primer = choose_cutoff_for_single_strand(
                input_fastq, cutoff_list, rev=True
            )  # .. assess and select

        # Output cutoff point(s) to .txt file
        with open(f"{output_prefix}_chosen_cutoffs.txt", "w") as fw:
            if f_cutoff != "":
                fw.write(f"F: {f_cutoff}\n")
            if r_cutoff != "":
                fw.write(f"R: {r_cutoff}\n")

        # Output consensus primer sequence(s) to .fasta file
        with open(f"{output_prefix}_auto_primers.fasta", "w") as fw:
            if f_cutoff != "":
                fw.write(f">F_auto\n{f_primer}\n")
            if r_cutoff != "":
                fw.write(f">R_auto\n{r_primer}\n")

    console.log(
        "[bold green]Choosing optimal primer cutoff point :white_check_mark:[/bold green]\n"
    )

    return Path(f"{output_prefix}_auto_primers.fasta")


@cli.command(
    "auto",
    options_metavar="-i <fastq/fastq.gz> -st [FR/F/R] -o <output_prefix>",
    short_help="Perform the primer cutoff strategy for primer inference",
)
@click.option(
    "-i",
    "--input_fastq",
    required=True,
    help="Input fastq file to perform primer inference on.",
    type=click.Path(exists=True, path_type=Path, dir_okay=False),
)
@click.option(
    "-st",
    "--strand",
    help="The strand(s) to perform primer inference for. Values can be either F, R, or FR for both.",
    type=click.Choice(["FR", "F", "R"]),
    required=True,
)
@click.option(
    "-o", "--output_prefix", required=True, help="Prefix to output file.", type=str
)
@click.pass_context
def primer_cutoff_strategy(
    ctx: click.Context,
    input_fastq: Path,
    strand: str,
    output_prefix: str,
) -> None:
    """Runs the primer cutoff strategy for primer inference.
    Using patterns of base-conservation values for input reads, this strategy identifies potential cutoff
    points where a primer sequence could end. An optimal choice of cutoff based on the change in base-conservation
    before and after a cutoff is then made. Finally a consensus sequence is generated using this cutoff point
    and the input reads, which are saved to an output FASTA file.

    :param ctx: Click context - needed to invoke the multiple Click functions this "pipeline" uses.
    :type ctx: click.Context
    :param input_fastq: The input FASTQ file.
    :type input_fastq: Path
    :param strand: The strand(s) to perform primer inference for. Values can be either F, R, or FR for both.
    :type strand: str
    :param output_prefix: The prefix to be used on output files.
    :type output_prefix: str
    """

    print(
        "[bold grey74]Running [bold green]primer cutoff strategy[/bold green].[/bold grey74]"
    )
    print(
        f"[bold grey74]Input FASTQ file: [bold green]{input_fastq}[/bold green][/bold grey74]"
    )
    print(
        f"[bold grey74]Strands to perform strategy on: [bold green]{strand}[/bold green][/bold grey74]"
    )
    print(
        f"[bold grey74]Output prefix: [bold green]{output_prefix}[/bold green][/bold grey74]"
    )
    print("")

    bcv_df = ctx.invoke(
        generate_base_conservation_vector,
        input_fastq=input_fastq,
        strand=strand,
        output_prefix=output_prefix,
    )

    cutoffs_path = ctx.invoke(
        find_potential_cutoffs, input_bcv=bcv_df, output_prefix=output_prefix
    )

    inferred_primer_seq_path = ctx.invoke(
        choose_primer_cutoff,
        input_fastq=input_fastq,
        primer_cutoffs=cutoffs_path,
        output_prefix=output_prefix,
    )

    print("[bold grey74]Primer cutoff strategy finished![/bold grey74]")
    primer_fasta = pyfastx.Fasta(str(inferred_primer_seq_path), build_index=False)
    print("[bold grey74]Inferred primers:[/bold grey74]")
    for primer in primer_fasta:
        print(f"[bold green]>{primer[0]}[/bold green]")
        print(f"[bold red]{primer[1]}[/bold red]")
    print("")
    print(
        f"[bold grey74]Final primers found saved to [bold green]{inferred_primer_seq_path}[/bold green][/bold grey74]"
    )


if __name__ == "__main__":
    cli()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2025 EMBL - European Bioinformatics Institute
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from pathlib import Path
import pyfastx

from Bio.Seq import Seq
import regex

from pimento.bin.pimento_utils import (
    primer_regex_query_builder,
    get_read_count,
    fetch_read_substrings,
)

from pimento.bin.thresholds import STD_PRIMER_READ_PREFIX_LENGTH, MAX_READ_COUNT


def parse_std_primers(
    primers_dir: Path,
    merged: bool = False,
) -> tuple[defaultdict[defaultdict], defaultdict[defaultdict]]:
    """
    Parse the library of standard primers.

    Reads the fasta files in the given directory
    Primer names (which are the fasta headers) are labeled with F or R for 5'-3' and 3'-5' primers respectively

    Returns two dictionaries:
        std_primer_dict_regex
            key: region+primer name
            val: primer sequence from 5' to 3'
        std_primer_dict
            key: region+primer name
            val: primer sequence from 5' to 3' for forward primers, 3' to 5' for reverse
    """

    std_primer_dict_regex = defaultdict(defaultdict)
    std_primer_dict = defaultdict(defaultdict)

    primer_files = list(primers_dir.glob("*.fasta"))

    primer_count = 0
    for primer_file in primer_files:
        region = primer_file.stem
        primer_fasta = pyfastx.Fasta(str(primer_file), build_index=False)

        for primer_name, primer_seq in primer_fasta:
            if merged and primer_name[-1] == "R":
                primer_seq = str(Seq(primer_seq).complement())

            primer_regex = primer_regex_query_builder(primer_seq)
            std_primer_dict_regex[region][primer_name] = primer_regex
            std_primer_dict[region][primer_name] = primer_seq
            primer_count += 1

    return std_primer_dict_regex, std_primer_dict, primer_count


def run_primer_matching_once(input_path: Path, input_primer: str, rev: bool = False):
    """
    Run primer matching using the regex package.

    Takes one primer, strand, and fastq input
    Uses fuzzy matching to allow for at most one error (for sequencing errors)
    Returns number of reads matching given primer
    """

    match_count = 0.0

    substring_count_dict = fetch_read_substrings(
        input_path, STD_PRIMER_READ_PREFIX_LENGTH, rev, max_line_count=MAX_READ_COUNT
    )

    for substring in substring_count_dict.keys():
        substring = substring.strip()
        res = regex.match(input_primer, substring)
        if res is not None:
            match_count += substring_count_dict[substring]

    return match_count


def get_primer_props(
    std_primer_dict_regex: defaultdict,
    input_fastq: Path,
    min_std_primer_threshold: float,
    merged: bool = False,
) -> list[str, dict]:
    """
    Look for the standard primers in the input fastq file.

    Will loop through the dictionary of primers, using fuzzy regex matching to find matching primers.
    If a std primer is present above a set threshold proportion, it is collected. Both strands are searched for.
    If there is an std primer for both the F and R strands, the maximum prop for each strand is chosen and the pair
    is output as a combination.

    Returns a list containing two elements:
        max_region: the amplified region the chosen primers belong to
        max_primers: dictionary containing the F and/or R primers that were chosen
    """

    read_count = get_read_count(
        input_fastq, file_type="fastq"
    )  # Get read count of fastq file to calculate proportion with
    res_dict = defaultdict(defaultdict)

    std_primer_log = open("all_standard_primer_proportions.txt", "w")

    # Loop through every primer region
    for region, primer in std_primer_dict_regex.items():
        res_dict[region]["F"] = {}
        res_dict[region]["R"] = {}

        # Loop through every primer of a certain region
        for primer_name, primer_seq in primer.items():

            region_name_str = f"{region};{primer_name}"
            primer_count = 0.0

            if merged and primer_name[-1] == "R":
                primer_count = run_primer_matching_once(
                    input_fastq, primer_seq, rev=True
                )  # Get count of a R primer with fuzzy regex matching
            else:
                primer_count = run_primer_matching_once(
                    input_fastq, primer_seq, rev=False
                )  # Get count of a F primer with fuzzy regex matching
            try:
                primer_prop = primer_count / read_count
            except ZeroDivisionError:
                primer_prop = 0

            if primer_name[-1] == "F":
                if (
                    primer_prop > min_std_primer_threshold
                ):  # Only collect primer if it's above threshold
                    res_dict[region]["F"][primer_name] = primer_prop
            elif primer_name[-1] == "R":
                if (
                    primer_prop > min_std_primer_threshold
                ):  # Only collect primer if it's above threshold
                    res_dict[region]["R"][primer_name] = primer_prop

            std_primer_log.write(f"{region_name_str}:{primer_prop}\n")

        # If an F or/and R primer wasn't found then just remove that key from the dictionary
        if res_dict[region]["F"] == {}:
            res_dict[region].pop("F")
        if res_dict[region]["R"] == {}:
            res_dict[region].pop("R")

    std_primer_log.close()

    singles = defaultdict(str)
    doubles = defaultdict(list)

    double_status = False  # Flag for whether primers were found on both strands

    #  Loop through every collected primer and put primers in singles or doubles
    for region in res_dict.keys():
        strands = res_dict[region]

        for strand in strands.keys():
            primers = strands[strand]
            max_prop = 0.0
            max_name = ""
            for primer_name, prop in primers.items():
                if prop > max_prop:
                    max_prop = prop
                    max_name = primer_name

            if len(strands.keys()) == 2:
                double_status = True
                doubles[region].append({max_name: max_prop})
            elif len(strands.keys()) == 1:
                singles[region] = {max_name: max_prop}

    max_region = ""
    max_primers = {}
    max_mean_prop = 0

    # if at least one pair of primers was collected
    if double_status:
        for (
            region
        ) in doubles:  # Loop through all pairs of primers and choose the best one
            primers = doubles[region]

            f_primer_name = list(primers[0].keys())[0]
            r_primer_name = list(primers[1].keys())[0]
            f_primer_prop = primers[0][f_primer_name]
            r_primer_prop = primers[1][r_primer_name]

            mean_prop = (f_primer_prop + r_primer_prop) / 2.0
            if mean_prop > max_mean_prop:
                max_mean_prop = mean_prop
                max_region = region
                max_primers = [
                    {f_primer_name: f_primer_prop},
                    {r_primer_name: r_primer_prop},
                ]

    else:
        for region in singles:  # Choose the best single primer
            primer = singles[region]
            primer_name = list(primer.keys())[0]
            prop = primer[primer_name]
            if prop > max_mean_prop:
                max_mean_prop = prop
                max_region = region
                max_primers = {primer_name: prop}

    if max_region == "":
        return []
    elif double_status:
        return [max_region, max_primers[0], max_primers[1]]
    else:
        return [max_region, max_primers]


def write_std_output(
    results: list[str, dict],
    output_prefix: str,
    std_primer_dict: defaultdict,
    merged: bool = False,
) -> None:
    """
    Save found std primers into a fasta file.
    """

    with (
        open(f"{output_prefix}_std_primers.fasta", "w") as fw_seq,
        open(f"{output_prefix}_std_primer_out.txt", "w") as fw_out,
    ):
        if results == []:  # if no primers found
            fw_out.write("")
            fw_seq.write("")

        elif len(results) == 2:  # if one primer found
            region = results[0]
            primer_name = list(results[1].keys())[0]
            primer_prop = results[1][list(results[1].keys())[0]]
            seq = std_primer_dict[region][primer_name]
            if merged and "R" in primer_name:
                seq = str(Seq(seq).complement())
            fw_out.write(f"{region}\n")
            fw_out.write(f"{primer_name}: {primer_prop}")

            fw_seq.write(f">{primer_name}\n{seq}")

        elif len(results) == 3:  # if primer pair found
            region = results[0]
            f_primer_name = list(results[1].keys())[0]
            f_primer_prop = results[1][list(results[1].keys())[0]]
            f_seq = std_primer_dict[region][f_primer_name]
            r_primer_name = list(results[2].keys())[0]
            r_primer_prop = results[2][list(results[2].keys())[0]]
            r_seq = std_primer_dict[region][r_primer_name]
            r_seq = str(Seq(r_seq).complement())

            fw_out.write(f"{region}\n")
            fw_out.write(f"{f_primer_name}: {f_primer_prop}\n")
            fw_out.write(f"{r_primer_name}: {r_primer_prop}")

            fw_seq.write(f">{f_primer_name}\n{f_seq}\n")
            fw_seq.write(f">{r_primer_name}\n{r_seq}\n")

    return Path(f"{output_prefix}_std_primers.fasta"), Path(
        f"{output_prefix}_std_primer_out.txt"
    )
